import numpy as np
import re

NL = "\n"

class ParseRST:
    
    blocks: list
    
    def __init__(self):
        self.blocks = []

    def title(self, level:int, text:str, doc:bool=False):
        underscore = "=-~^"[level]
        if doc:
            text = underscore*len(text) + NL + text + NL + underscore*len(text)
        else:
            text = text + NL + underscore*len(text)
        self.blocks.append(text)
        
    def paragraph(self, text:str):
        #text = NL.join([line.strip() for line in text.split(NL)])
        self.blocks.append(text)
        
    def table_csv(self, data, **kwargs):
        text = ".. csv-table::"
        for key, value in kwargs.items():
            text += f"   :{key}: {value}" + NL
        text += NL+NL
        for row in data:
            text += "   " + ", ".join([f'"{col}"' for col in row]) + NL
        self.blocks.append(text)
        
    def table(self, data, **kwargs):
        # Calculate maximum column widths
        widths = np.zeros_like(data, dtype=int)
        for r,row in enumerate(data):
            widths[r] = [len(col)+2 for col in row]
        widths = widths.max(axis=0)
        # Construct table matrix
        shape = np.array(data).shape
        table = np.full((shape[0]*2+1, shape[1]*2+1), '', dtype=object)
        for r,row in enumerate(data):
            for c,col in enumerate(row):
                m, n = 1+r*2, 1+c*2
                table[m, n] = f" {col:{widths[c]-1}s}"
        # Create borders
        hcol = 1 if 'header' in kwargs and kwargs['header'] else -1
        for r,row in enumerate(data):
            for c,col in enumerate(row):
                ls = "=" if r==hcol else "-"
                table[r*2, 1+c*2] = ls*widths[c]
                table[r*2, c*2] = "+"
                table[1+r*2, c*2] = "|"
            table[r*2,-1] = "+"
            table[1+r*2,-1] = "|"
        r += 1
        for c,col in enumerate(row):
            table[r*2, 1+c*2] = '-'*widths[c]
            table[r*2, c*2] = "+"
        table[r*2,-1] = "+"
        # Merge cells
        if 'merge' in kwargs:
            for r0,c0,r1,c1 in kwargs['merge']:
                for r in range(1+r0*2,2+r1*2):
                    for c in range(1+c0*2,2+c1*2):
                        table[r,c] = ''
                w = sum([widths[w] for w in range(c0,1+c1)])
                w += c1-c0-1
                table[1+r0*2,1+c0*2] = f" {data[r0][c0]:{w}s}"
        # Create table text
        text = table.sum(axis=1)
        text = NL.join(text)
        self.blocks.append(text)
    
    def target(self, key:str):
        text = f".. _{key}:"
        self.blocks.append(text)
    
    def link(self, key:str, name:str, source:bool=False):
        if source:
            key = re.sub('_[0-9]+$', '', key)
        return f":ref:`{name}<{key}>`" 

    def code(self, code:str, language:str='', **kwargs):
        text = f".. code-block:: {language}" + NL
        for key, value in kwargs.items():
            if key in ['linenos']:
                text += f"   :{key}:" + NL
            else:
                text += f"   :{key}: {str(value)}" + NL
        text += NL
        text += "   "+code.replace(NL,f"{NL}   ")
        self.blocks.append(text)

    def toc(self, pages:list, **kwargs):
        text = ".. toctree::" + NL
        for key, value in kwargs.items():
            text += f"   :{key}: {str(value)}" + NL
        text += NL
        for page in pages:
            text += f"   {page}" + NL
        self.blocks.append(text)
        
    def parse(self):
        return (NL+NL).join(self.blocks)


from .settings import *

class UnitItem:
    target: str
    link_source: str
    name: str
    value: str
    units: str
    source: tuple

    def __init__(self, name, unit):
        self.target = UnitItemTarget(name)
        self.link_source = SourceItemTarget(unit['source'][0],unit['source'][1])
        self.name   = name
        self.value  = unit['value']
        self.units  = unit['units']
        self.source = unit['source']



import numpy as np
import re

from .substance import Substance
from ..solver import *

class MaterialSolver:
    
    atom
    
    def __init__(self, atom:callable):
        self.atom = atom

    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass

    def preprocess(self, expr):
        PATTERN_NUMBER = "([0-9]+(\.([0-9]+([eE]{1}[+-]?[0-9]{0,3}|)|)|))"
        PATTERN_MOLECULE = "(\<[^\<\>]+\>)"
        def repl1(m):
            return m.group(1)+CustomOperatorMul.symbol+m.group(5)
        expr = re.sub(PATTERN_NUMBER+"\s*"+PATTERN_MOLECULE, repl1, expr)
        def repl1(m):
            return m.group(1)+CustomOperatorAdd.symbol+m.group(2)
        expr = re.sub(PATTERN_MOLECULE+"\s*"+PATTERN_NUMBER, repl1, expr)
        def repl1(m):
            return m.group(1)+CustomOperatorAdd.symbol+m.group(2)
        expr = re.sub(PATTERN_MOLECULE+"\s*"+PATTERN_MOLECULE, repl1, expr)
        return expr

    def solve(self, expr):
        expr_bak = expr

        # preprocess expression
        expr = self.preprocess(expr)
        # solve expression
        operators = {
            'par':CustomOperatorPar,  # should be the last of parenthesis operators
            'mul':CustomOperatorMul,
            'add':CustomOperatorAdd, 
        }
        steps = [
            dict(operators=['par'],  otype=Otype.ARGS),
            dict(operators=['mul'],  otype=Otype.BINARY),
            dict(operators=['add'],  otype=Otype.BINARY),
        ]
        with ExpressionSolver(self.atom, operators, steps) as es:
            material = es.solve(expr)
            
        return material

class CustomOperatorPar(OperatorPar):
    symbol: str = '<'            # operator specifc opening symbol
    symbol_open: str = '<'       # general opening symbol
    symbol_separator: str = ','  # argument separator
    symbol_close: str = '>'      # general closing symbol 

class CustomOperatorAdd(OperatorAdd):
    symbol: str = ' + '

class CustomOperatorMul(OperatorMul):
    symbol: str = ' * '


import numpy as np
import os
import sys
sys.path.insert(0, os.environ['DIR_SOURCE'])

from scinumtools import RowCollector
from scinumtools.units import Dimensions
from scinumtools.units.settings import *
from scinumtools.units.unit_types import *
from scinumtools.dip import DIP
from scinumtools.dip.config import *

dir_export = os.environ['DIR_DOCS']+"/source/_static/export_config"
file_code = f"{dir_export}/config.dip"

def build_config_c():
    
    file_export = f"{dir_export}/config_c.h"
    print(file_export)
    
    with DIP() as dip:
       dip.add_file(file_code)
       env = dip.parse()
    with ExportConfigC(env) as exp:
        exp.parse(
            guard='CONFIG_H',
            define=['radiation','simulation.name']
        )
        exp.save(file_export)
        
def build_config_cpp():
    
    file_export = f"{dir_export}/config_cpp.h"
    print(file_export)
    
    with DIP() as dip:
       dip.add_file(file_code)
       env = dip.parse()
    with ExportConfigCPP(env) as exp:
        exp.parse(
            guard='CONFIG_H',
            define=['radiation','simulation.name'],
            const=['box.width','box.height']
        )
        exp.save(file_export)
        
def build_config_rust():
    
    file_export = f"{dir_export}/config_rust.rs"
    print(file_export)
    
    with DIP() as dip:
       dip.add_file(file_code)
       env = dip.parse()
    with ExportConfigRust(env) as exp:
        exp.parse()
        exp.save(file_export)

def build_config_fortran():
    
    file_export = f"{dir_export}/config_fortran.f90"
    print(file_export)
    
    with DIP() as dip:
       dip.add_file(file_code)
       env = dip.parse()
    with ExportConfigFortran(env) as exp:
        exp.parse()
        exp.save(file_export)
        
def build_config_bash():
    
    file_export = f"{dir_export}/config_bash.sh"
    print(file_export)
    
    with DIP() as dip:
       dip.add_file(file_code)
       env = dip.parse()
    with ExportConfigBash(env) as exp:
        exp.parse()
        exp.save(file_export)
        
def build_config_yaml():
    
    file_export = f"{dir_export}/config_yaml.yaml"
    print(file_export)
    
    with DIP() as dip:
       dip.add_file(file_code)
       env = dip.parse()
    with ExportConfigYAML(env) as exp:
        exp.parse(units=True, default_flow_style=False)
        exp.save(file_export)
        
def build_config_toml():
    
    file_export = f"{dir_export}/config_toml.toml"
    print(file_export)
    
    with DIP() as dip:
       dip.add_file(file_code)
       env = dip.parse()
    with ExportConfigTOML(env) as exp:
        exp.parse()
        exp.save(file_export)

def build_config_json():
    
    file_export = f"{dir_export}/config_json.json"
    print(file_export)
    
    with DIP() as dip:
       dip.add_file(file_code)
       env = dip.parse()
    with ExportConfigJSON(env) as exp:
        exp.parse(indent=2)
        exp.save(file_export)


from typing import List, Dict, Union
from dataclasses import dataclass, field

from ..settings import Sign
from ...units import Quantity, UnitEnvironment

@dataclass
class UnitList:
    units: Dict   = field(default_factory = dict)  # custom units

    def __len__(self):
        return len(self.units)
    
    def __getitem__(self, key):
        return self.units[key]
    
    def items(self):
        return self.units.items()
    
    def keys(self):
        return self.units.keys()
        
    def append(self, name:str, value:str, units:str, unit:Quantity, source:tuple):
        """ Add a new source

        :param str name: Name of a new unit
        :param Quantity unit: Quantity object
        """
        name = f"[{name}]"
        if name in self.units:
            raise Exception("Reference unit alread exists:", name)
        self.units[name] = {
            'magnitude':  unit.magnitude.value, 
            'dimensions': unit.baseunits.dimensions.value(),
            'value': value, 
            'units': units,
            'source': source,
        } #,'prefixes':['k','M','G']}
        
    def query(self, query:str):
        """ Select units according to a query

        :param str query: unit selection query
        """
        if query==Sign.WILDCARD:   # return all units
            return self.units
        else:                     # return particular unit
            if query not in self.units:
                raise Exception("Requested unit does not exists:", query)
            return {query: self.units[query]}


import numpy as np
import json
import re

from .node import Node
from ..datatypes import Type, NumberType, IntegerType, FloatType, StringType, BooleanType
from ..environment import Environment
from ..settings import Keyword, Sign, EnvType

class BaseNode(Node):
    
    value: str = None       # Value
    constant: bool = False
    condition: str = None
    branch_id: int = None
    case_id: int = None
    
    def __init__(self, parser=None, *args, **kwargs):
        if parser:
            kwargs['code'] = parser.code
            kwargs['source'] = parser.source
            kwargs['indent'] = parser.indent
            kwargs['name'] = parser.name
            kwargs['dtype_prop'] = parser.dtype_prop
            kwargs['value_raw'] = parser.value_raw
            kwargs['value_ref'] = parser.value_ref
            kwargs['value_slice'] = parser.value_slice
            kwargs['units_raw'] = parser.units_raw
            kwargs['dimension'] = parser.dimension
            kwargs['defined'] = parser.defined
            if parser.keyword:
                kwargs['keyword'] = parser.keyword
            if parser.value_fn:
                kwargs['value_fn'] = parser.value_fn
            if parser.value_expr:
                kwargs['value_expr'] = parser.value_expr
        super().__init__(*args, **kwargs)
    
    def __str__(self):
        class_name = self.__class__.__name__.replace('Node','')
        node_name = (self.name.split('.'))[-1]
        if isinstance(self.value, NumberType) and self.value.unit:
            return f"{class_name}({node_name}: {self.value.value} {self.value.unit})"
        elif isinstance(self.value, Type):
            return f"{class_name}({node_name}: {self.value.value})"
        else:
            return f"{class_name}({node_name}: {self.value})"
    
    def __repr__(self):
        class_name = self.__class__.__name__.replace('Node','')
        node_name = (self.name.split('.'))[-1]
        if isinstance(self.value, NumberType) and self.value.unit:
            return f"{class_name}({node_name}: {self.value.value} {self.value.unit})"
        elif isinstance(self.value, Type):
            return f"{class_name}({node_name}: {self.value.value})"
        else:
            return f"{class_name}({node_name}: {self.value})"

    def parse(self, env):
        return False

    def clean_name(self):
        """ Strip @case and @else from the name
        """
        pattern = f"{Sign.CONDITION}[0-9]+{Sign.SEPARATOR}"
        return re.sub(pattern, '', self.name)

    def cast_value(self, value=None):
        """ Cast (raw-)value as a datatype self, or another node
        """
        if not value:
            if self.value is None:
                value = self.value_raw
            else:
                if isinstance(self.value, (StringType, BooleanType, NumberType)):
                    value = self.value.value
                else:
                    value = self.value
        if np.isscalar(value) and value in [None, Keyword.NONE]:
            value = None
        elif self.dimension or self.value_slice:
            # cast multidimensional values
            if isinstance(value, str):
                value = np.array(json.loads(value), dtype=self.dtype)
            else:
                value = np.array(value, dtype=self.dtype)
            if self.value_slice:
                value = self.slice_value(self.value_slice, value)
            if self.dimension:
                # check if dimensions are correct
                for d,dim in enumerate(self.dimension):
                    shape = value.shape[d]
                    if dim[0] is not None and shape < dim[0]:
                        raise Exception(f"Node '{self.name}' has invalid dimension: dim({d})={shape} < {dim[0]}")
                    if dim[1] is not None and shape > dim[1]:
                        raise Exception(f"Node '{self.name}' has invalid dimension: dim({d})={shape} > {dim[1]}")
            else:
                if not np.isscalar(value):
                    raise Exception("Array value set to scalar node:",self.code,value)
        else:
            # cast scalar values
            if value is not None:
                # casting string as boolean returns true always if string is non-empty
                # that's why we need to convert it expicitely
                if self.keyword=='bool':
                    if isinstance(value, BooleanType):
                        pass
                    elif isinstance(value, (bool,np.bool_)):
                        value = BooleanType(value)
                    elif value==Keyword.TRUE:
                        value = BooleanType(True)
                    elif value==Keyword.FALSE:
                        value = BooleanType(False)
                    else:
                        raise Exception("Could not convert raw value to boolean type:",value)
                else:
                    try:
                        if isinstance(value, FloatType):
                            if self.keyword=='int':
                                value = IntegerType(value.number, value.unit)
                        elif isinstance(value, IntegerType):
                            if self.keyword=='float':
                                value = FloatType(value.number, value.unit)
                        else:
                            value = self.dtype(value)
                    except:
                        raise Exception("Could not convert raw value to type:",self.code,value)
        return value

    def set_value(self, value=None):
        """ Set value using value_raw or arbitrary value
        """
        if value is None and self.value_raw:
            self.value = self.cast_value()
        elif value:
            self.value = value
        else:
            self.value = None
    
    def modify_value(self, node, env):
        """ Modify value taking value of a different node
        """
        if node.keyword!='mod' and node.dtype!=self.dtype:
            raise Exception(f"Datatype {self.dtype} of node '{self.name}' cannot be changed to {node.dtype}")
        if not self.value:  # create a dummy value if none
            self.set_value(node.value_raw)
        # copy value type modify values and units
        value = self.value.copy()
        value.value = self.cast_value(node.value_raw)
        if isinstance(value, (IntegerType, FloatType)):
            value.unit = node.units_raw
            value.convert(self.units_raw, env)
        self.set_value(value.value)

    def slice_value(self, slices, value=None):
        """ Slice part of the value

        :param list slices: List of tuples with slicing
        :param value: Value to be sliced. If none value of current node is taken.
        """
        if value is None:
            value = self.value
        if isinstance(value, Type):
            value = value.value
        smin, smax = slices.pop(0)
        if smin==smax and smin is not None:
            value = value[smin]
        elif smin!=smax:
            value = value[slice(smin,smax)]                  
        if slices:
            if smin==smax and smin is not None:
                return self.slice_value(slices.copy(), value)
            else:
                return np.array([self.slice_value(slices.copy(), val) for val in value])
        else:
            return value

    def inject_value(self, env:Environment, node=None):
        """ Inject value from a remote source

        :param env: Environment object
        :param node: Node object that should be injected
        """
        if not node:
            node = self
        if not node.value_ref:
            return
        if env.envtype==EnvType.DOCS:
            nodes = env.request(node.value_ref, count=[0,1], errsrc=False)
            if len(nodes)==0:
                return
        else:
            nodes = env.request(node.value_ref, count=1)
        if isinstance(nodes, str):   # block import
            node.value_raw = nodes
        else:                        # node import
            node.value_raw = nodes[0].value_raw
            if not node.units_raw:
                node.units_raw = nodes[0].units_raw
        



import numpy as np
import os
import sys
sys.path.insert(0, os.environ['DIR_SOURCE'])

from scinumtools import RowCollector
from scinumtools.units import Dimensions
from scinumtools.units.settings import *
from scinumtools.units.unit_types import *
from scinumtools.dip import DIP
from scinumtools.dip.docs import *

def build_docs_pdf():
    
    # Generate a PDF documentation from a DIP file
    dir_pdf = os.environ['DIR_DOCS']+"/source/_static/pdf"
    file_definitions = f"{dir_pdf}/definitions.dip"
    file_cells = f"{dir_pdf}/cells.dip"
    file_pdf = f"{dir_pdf}/documentation.pdf"
    # Create directory if missing
    if not os.path.isdir(dir_pdf):
        os.mkdir(dir_pdf)
    # Create a DIP environment
    with DIP(name="PDF") as p:
        p.add_unit("velocity", 13, 'cm/s')
        p.add_string("""
        $unit length = 1 cm
        $unit mass = 2 g
        
        cfl_factor float = 0.7  # Courant–Friedrichs–Lewy condition
        max_vare float = 0.2    # maximum energy change of electrons
        max_vari float = 0.2    # maximum energy change of ions
        """)
        p.add_source("cells", file_cells)
        p.add_file(file_definitions)
        docs = p.parse_docs()    # Export parameters as a PDF
    with ExportDocsPDF(docs) as exp:
        exp.build(
            file_pdf, 
            "Example DIP documentation", 
            """In this document we want to demonstrate basic capabilities of a DIP documentation.<br/><br/>
            The documentation is structured into 3 main sections. The first section summarizes all parameters
            in a DIP code, as well as their corresponding node definitions, declarations, modifications and corresponding properties.
            Following section summarizes all references of injected values and lists imported nodes.
            The final section gives an overview of custom units and code sources.<br/><br/>
            Parameters, nodes, sections and many other items in this documentation are cross-linked between
            each other. All hyperlinks are denoted with a blue text.
        """)
    print(file_pdf)

def build_docs_html():
    
    # Generate a PDF documentation from a DIP file
    dir_docs = os.environ['DIR_DOCS']+"/source/_static/htmldocs"
    dir_docs_build = f"{dir_docs}/build"
    file_definitions = f"{dir_docs}/definitions.dip"
    file_cells = f"{dir_docs}/cells.dip"
    file_html = f"{dir_docs}/documentation.pdf"
    # Create directory if missing
    if not os.path.isdir(dir_docs_build):
        os.mkdir(dir_docs_build)
    if not os.path.isdir(dir_docs):
        os.mkdir(dir_docs)
    # Create a DIP environment
    with DIP(name="PDF") as p:
        p.add_unit("velocity", 13, 'cm/s')
        p.add_string("""
        $unit length = 1 cm
        $unit mass = 2 g
        
        cfl_factor float = 0.7  # Courant-Friedrichs-Lewy condition
        max_vare float = 0.2    # maximum energy change of electrons
        max_vari float = 0.2    # maximum energy change of ions
        """)
        p.add_source("cells", file_cells)
        p.add_file(file_definitions)
        docs = p.parse_docs()    # Export parameters as a PDF
    with ExportDocsHTML(docs) as exp:
        exp.build(
            dir_docs_build, 
            "Example DIP documentation", 
            """In this document we want to demonstrate basic capabilities of a DIP documentation.<br/><br/>
            The documentation is structured into 3 main sections. The first section summarizes all parameters
            in a DIP code, as well as their corresponding node definitions, declarations, modifications and corresponding properties.
            Following section summarizes all references of injected values and lists imported nodes.
            The final section gives an overview of custom units and code sources.<br/><br/>
            Parameters, nodes, sections and many other items in this documentation are cross-linked between
            each other. All hyperlinks are denoted with a blue text.
        """)
    print(dir_docs_build)

def build_docs_rst():
    
    # Generate a PDF documentation from a DIP file
    dir_docs = os.environ['DIR_DOCS']+"/source/dip/docs/rst"
    dir_docs_build = f"{dir_docs}/build"
    file_definitions = f"{dir_docs}/definitions.dip"
    file_cells = f"{dir_docs}/cells.dip"
    file_html = f"{dir_docs}/documentation.pdf"
    # Create directory if missing
    if not os.path.isdir(dir_docs_build):
        os.mkdir(dir_docs_build)
    if not os.path.isdir(dir_docs):
        os.mkdir(dir_docs)
    # Create a DIP environment
    with DIP(name="PDF") as p:
        p.add_unit("velocity", 13, 'cm/s')
        p.add_string("""
        $unit length = 1 cm
        $unit mass = 2 g
        
        cfl_factor float = 0.7  # Courant-Friedrichs-Lewy condition
        max_vare float = 0.2    # maximum energy change of electrons
        max_vari float = 0.2    # maximum energy change of ions
        """)
        p.add_source("cells", file_cells)
        p.add_file(file_definitions)
        docs = p.parse_docs()    # Export parameters as a PDF
    with ExportDocsRST(docs) as exp:
        exp.build(
            dir_docs_build, 
            "RST documentation", """
DIP package comes with an automatized RST export of parameters.
As an example, we use the same files as in HTML and PDF documentations.

Environment suitable for a documentation has to be parsed with a special method ``parse_docs()``, that processes node differently as the standard ``parse()`` method.

.. code-block:: python

   >>> from scinumtools.dip import DIP
   >>> from scinumtools.dip.docs import ExportDocsRST
   >>> with DIP() as p:
   >>>     p.add_file('definitions.dip')
   >>>     docs = p.parse_docs()
   >>> with ExportDocsRST(docs) as exp:
   >>>     exp.build(
   >>>         './build', 
   >>>         "Example DIP documentation", 
   >>>         "In this document we want to demonstrate basic capabilities of a DIP documentation..... "
   >>>     )

``ExportDocsRST`` class provided above can be used as it is, or as a template for your own personalized documentation.
When building your own documentation, you can simply take the `existing source code <https://github.com/vrtulka23/scinumtools/tree/main/src/scinumtools/dip/docs/rst>`_ and modify it according to your needs.
RST code is generated using a custom `RST parser class <https://github.com/vrtulka23/scinumtools/tree/main/src/scinumtools/dip/docs/rst/rst_parser.py>`_.

Code above will generate the following default RST documentation.
        """)
    print(dir_docs_build)


import os
from pathlib import Path
import numpy as np
from inspect import getframeinfo, stack

from ..environment import Environment
from ..nodes.parser import Parser
from ..datatypes import Type

class TemplateSolver:

    env: Environment
    filename: str

    def __init__(self, env:Environment=None, **kwargs):
        if env:
            self.env = env
        else:
            self.env = Environment()
        if 'source' not in kwargs:
            # determine which file instantiate this class
            caller = getframeinfo(stack()[1][0])
            self.filename = caller.filename
        else:
            source = env.sources[kwargs['source'][0]]
            self.filename = source.path
            
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
    
    def solve(self, expr):
        out = ''
        kwargs = {'keyword':'expr'}
        while len(expr)>0:
            sign, expr = expr[0], expr[1:]
            if sign=='{':
                p = Parser(code=expr, **kwargs)
                p.part_reference()
                p.part_slice()
                p.part_format()
                if p.value_ref and p.ccode[0]=='}':
                    nodes = self.env.request(p.value_ref, count=1)
                    if p.value_slice:
                        value = nodes[0].slice_value(p.value_slice)
                    else:
                        value = nodes[0].value
                    if isinstance(value, Type):
                        value = value.value
                    if p.formating:
                        form = "{0"+p.formating+"}"
                        out += form.format(value)
                    else:
                        out += str(value)
                    expr = p.ccode[1:]
                else:
                    out += sign
            else:
                out += sign
        return out

    def template(self, file_in, file_out=None):
        if not os.path.isabs(file_in):
            # set relative paths with respect to the source script
            parent = Path(self.filename).parent
            file_in = parent / file_in
        with open(file_in,'r') as f:
            template = f.read()
        text = self.solve(template)
        if file_out:
            if not os.path.isabs(file_out):
                # set relative paths with respect to the source script
                parent = Path(self.filename).parent
                file_out = parent / file_out
            with open(file_out,'w') as f:
                f.write(text)
        return text
            



import numpy as np
import re

from .element import Element
from ..solver import *

class SubstanceSolver:
    
    atom
    
    def __init__(self, atom:callable):
        self.atom = atom

    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass

    def preprocess(self, expr):
        pattern = "(([A-Z]+[a-z]?|\[p\]|\[n\]|\[e\])(\{[0-9+-]+\}|)([0-9]*))"
        # insert implicit additions
        while True:
            expr_new = re.sub(pattern+"\s*"+pattern,"\g<1> + \g<5>",expr,count=1)
            if expr_new == expr:
                break
            else:
                expr = expr_new
        # insert implicit multiplications
        def repl1(m):
            if m.group(4):
                return m.group(2)+m.group(3)+CustomOperatorMul.symbol+m.group(4)
            else:
                return m.group(2)+m.group(3)
        expr = re.sub(pattern, repl1, expr)
        # insert implicit operation around parentheses
        def repl2(m):
            if m.group(1):
                return m.group(1) + CustomOperatorAdd.symbol + "("
            else:
                return m.group(2) + "("
        expr = re.sub("([^*+(\s]*)(\s*)\(", repl2, expr)
        def repl3(m):
            if m.group(1) and m.group(3):
                return ")" + CustomOperatorMul.symbol + m.group(1) + CustomOperatorAdd.symbol + m.group(3)
            elif m.group(1):
                return ")" + CustomOperatorMul.symbol + m.group(1)
            elif m.group(3):
                return ")" + CustomOperatorAdd.symbol + m.group(3)
            else:
                return ")"+m.group(2)
        expr = re.sub("\)([0-9]*)(\s*)([^+*)\s]*)", repl3, expr)
        return expr

    def solve(self, expr):
        expr_bak = expr

        # preprocess expression
        expr = self.preprocess(expr)
        # solve expression
        operators = {
            'par':OperatorPar,  # should be the last of parenthesis operators
            'mul':CustomOperatorMul,
            'add':CustomOperatorAdd, 
        }
        steps = [
            dict(operators=['par'],  otype=Otype.ARGS),
            dict(operators=['mul'],  otype=Otype.BINARY),
            dict(operators=['add'],  otype=Otype.BINARY),
        ]
        with ExpressionSolver(self.atom, operators, steps) as es:
            material = es.solve(expr)
            
        return material

class CustomOperatorAdd(OperatorAdd):
    symbol: str = ' + '

class CustomOperatorMul(OperatorMul):
    symbol: str = ' * '


import sys, os
import pytest
import numpy as np
sys.path.insert(0, 'src')

from scinumtools.dip import DIP
from scinumtools.dip.settings import Format
from scinumtools.dip.datatypes import FloatType, IntegerType

def parse(code):
    with DIP() as p:
        p.add_string(code)
        return p.parse().data(verbose=True,format=Format.TYPE)

def test_hierarchy():
    data = parse('''
general.colonel int = 1  # namespace notation
  captain                # group nodes
     soldier int = 2     # lowest node in the hierarchy
    ''')
    np.testing.assert_equal(data,{
        'general.colonel': IntegerType(1),
        'general.colonel.captain.soldier': IntegerType(2)
    })

def test_modification():
    data = parse('''
size float = 70 cm    # definition
size float = 80 cm    # modifications of value
size = 90 cm          # omitting datatype
size = 100            # omitting units
size = 1 m            # using different prefix

energy float = 1.23 J # definition
energy = 2.2 erg      # switching from SI to cgs
energy = 2.2 g*cm2/s2 # using unit expressions

angle float = 1.57079633 rad  # definition in radians
angle = 31 '                  # angle minutes

alcohol float = 34 %  # definition
alcohol = 2 ppth      # converting dimensionless units

temp float = 20 Cel
temp = 280.15 K
    ''')
    np.testing.assert_equal(data,{
        'size':   FloatType(100, 'cm'),
        'energy': FloatType(2.2000000000000004e-07, 'J'),
        'angle':  FloatType(0.0090175342, 'rad'),
        'alcohol': FloatType(0.2, '%'),
        'temp': FloatType(7, 'Cel'),
    })  
    
def test_modification_errors():
    
    with pytest.raises(Exception) as e_info:
        parse('''
age int = 34 yr
age float = 55
        ''')
    assert e_info.value.args[0] == "Datatype <class 'int'> of node 'age' cannot be changed to <class 'float'>"
    with pytest.raises(Exception) as e_info:
        parse('''
weight = 23 kg
        ''')
    assert e_info.value.args[0] == "Modifying undefined node:"
    assert e_info.value.args[1] == "weight"

def test_option_units():
    data = parse("""
width float = 2 m
  = 2 m
  = 3 m
width = 3000 mm
    """)
    np.testing.assert_equal(data,{
        'width': FloatType(3.0, 'm')
    })
    with pytest.raises(Exception) as e_info:
        parse('''
size float = 24 cm
  = 24 cm
  = 25 m
size = 25 cm
        ''')
    assert e_info.value.args[0] == "Value 'FloatType(25.0 cm)' of node 'size' doesn't match with any option:"
    assert e_info.value.args[1] == [FloatType(24.0, 'cm'), FloatType(2500.0, 'cm')]
    
if __name__ == "__main__":
    # Specify wich test to run
    test = sys.argv[1] if len(sys.argv)>1 else True

    # Loop through all tests
    for fn in dir(sys.modules[__name__]):
        if fn[:5]=='test_' and (test is True or test==fn[5:]):
            print(f"\nTesting: {fn}\n")
            locals()[fn]()
